;; -*- lexical-binding: t; -*-

;; Copyright (C) 2016-2022  Free Software Foundation, Inc.

(ert-deftest assess-discover-test-has-this-been-defined ()
  "This test is here so that we can test elsewhere that is has
actually been defined."
    (should t))

(provide 'assess-discover-test)
