;; -*- lexical-binding: t; -*-

;; Copyright (C) 2016-2022  Free Software Foundation, Inc.

(defun functionname (_args &optional _body)
  (+ 1 2))

(defun functionname2 (_args &optional _body)
  (+ 1 3))
