;; Generated package description from assess.el  -*- no-byte-compile: t -*-
(define-package "assess" "0.6.0.20231120.170108" "Test support functions" '((emacs "24.4") (m-buffer "0.15")) :commit "2399936d5b75094ed0e131b3f9d48041005383e2" :url "https://elpa.gnu.org/packages/assess.html" :authors '(("Phillip Lord" . "phillip.lord@russet.org.uk")) :maintainer '("Phillip Lord" . "phillip.lord@russet.org.uk"))
